import numpy as np
import matplotlib.pyplot as plt
from pde import *
from droplets import *
from droplet_cavitation import *


# Simulation parameters
grid_size = 200  # Squared grid is used
Nx = 8  # Number of support points along each grid dimensions
n_droplets = 100  # Number of droplets
t_simulation = 1000  # Final simulation type in adimentionalized units

## Linear stopping rate
cr = 10.0  # cooling rate in centigrades per second
alpha = 0.0932 / 100  # change in equilibrium concentration per centigrade
t_adim = 0.02  # adimentionalization parameter for time

# setup dilute concentration field
grid = CartesianGrid([[0, grid_size]] * 3, Nx, periodic=True)
background = DiffusionBackground()

c0 = 0.06939
# range of concentrations at which the mesh breaks
ceq_el_local_range = (1.0, 57.25)
# mesh size of the droplets
mesh_size_range = (0.1, 0.1)
mean_mesh_size = 0.1


# distribution of the cavitation pressures
height_dist = np.random.uniform(
    low=ceq_el_local_range[0], high=ceq_el_local_range[1], size=n_droplets
)

lambda_break = 10  # at what stretch ratio the mesh breaks
Ym = 186  # young modulus in kilopascals
# define the pressure curve
outside_field = SingleQuench(
    cr, c_initial=c0, t_adim=t_adim, lambda_break=lambda_break, Ym=Ym
)

# initialize droplets with a radius equivalent to the mesh size
droplets = [
    ElasticDroplet(
        grid.get_random_point(), mean_mesh_size, mesh_size_range[0], height_dist[_]
    )
    for _ in range(n_droplets)
]


parameters = {"equilibrium_concentration": outside_field}
agents = HeterogenousPointAgents(parameters)

simulation = AgentSimulation(background, agents)
# setup simulation state
state = SimulationState(
    background=ScalarBackgroundState(grid, c0),
    agents=HeterogenousAgentsState(droplets=droplets),
)


# run simulation
interval = 20  # tracking interval
tracker = DropletAgentTracker(interval)
result = simulation.run(state, t_range=t_simulation, tracker=["progress", tracker])


# Plot of droplet radius over time
plt.rcParams["figure.figsize"] = (10, 7)
em = tracker.emulsions[-1]
tracker.droplet_tracks.plot(attribute="radius", color="k", alpha=0.1)
if em.data is not None:
    plt.title(
        f"{sum(em.data['radius'] > 2)} droplets with $R>2$ from {result.agents.droplet_count} droplets"
    )
plt.xlabel("t (AU)")
plt.ylabel("Radius (AU)")
