'''
Module defining the abstract base class for agents

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

from typing import Dict, Any, Type, Callable  # @UnusedImport
import logging
from abc import abstractmethod, ABCMeta
import json

import numpy as np

from pde.tools.parameters import Parameterized
from pde.tools.cache import objects_equal
from droplets import Emulsion

from ..backgrounds.base import BackgroundStateBase



class AgentsStateBase(Parameterized):
    """ represents the state of many agents of the same type """
    
    data: np.ndarray
    """ :class:`~numpy.ndarray`:
    Data describing the agents. These are the dynamical variables (degree of
    freedoms) of the simulation
    """
    
    droplets: Emulsion
    """ :class:`droplets.emulsions.Emulsion`:
    Represents the agents in form of an Emulsion for further processing
    """
    
    _subclasses: Dict[str, 'AgentsStateBase'] = {}  # type: ignore


    @classmethod
    def from_state(cls, attributes: Dict[str, Any], data: np.ndarray) \
            -> "AgentsStateBase":
        """ create the agents state from attributes and data
        
        Args:
            attributes (dict):
                Attributes of the agents state
            data (:class:`~numpy.ndarray`):
                The numerical data associated with the agents
        """
        raise NotImplementedError

   
    @classmethod
    def from_hdf_dataset(cls, dataset) -> "AgentsStateBase":
        """ construct the agents state by reading data from an hdf5 dataset
        
        Args:
            dataset: the hdf5 dataset (in an already opened file)
        """
        # copy attributes from hdf
        attributes = dict(dataset.attrs)
        
        # determine class
        class_name = json.loads(attributes.pop('class'))
        field_cls = cls._subclasses[class_name]
        
        # unserialize the attributes
        attributes = cls.unserialize_attributes(attributes)

        # construct the instance
        return field_cls.from_state(attributes, data=dataset)

    
    def __str__(self):
        return f"{self.__class__.__name__}(<{self.droplet_count} droplets>)"
    __repr__ = __str__


    @property
    def attributes(self) -> Dict[str, Any]:
        """ dict: information about the agents state """
        return {'class': self.__class__.__name__,
                'parameters': self.parameters}


    @property
    def attributes_serialized(self) -> Dict[str, str]:
        """ dict: serialized version of the attributes """
        return {key: json.dumps(value)
                for key, value in self.attributes.items()}
    

    @classmethod
    def unserialize_attributes(cls, attributes: Dict[str, str]) \
            -> Dict[str, Any]:
        """ unserializes the given attributes
        
        Args:
            attributes (dict):
                The serialized attributes
                
        Returns:
            dict: The unserialized attributes
        """
        return {key: json.loads(value)
                for key, value in attributes.items()}


    def to_file(self, filename: str, **kwargs):
        r""" store agents state in a file
        
        Args:
            filename (str):
                Path where the data is stored
            \**kwargs:
                Additional parameters may be supported for some formats 
        """
        import h5py
        with h5py.File(filename, "w") as fp:
            self._write_hdf_dataset(fp, **kwargs)
                    
    
    def _write_hdf_dataset(self, hdf_path, key: str = 'data'):
        """ write data to a given hdf5 file pointer `hdf_path` """
        dataset = hdf_path.create_dataset(key, data=self.data)
        
        # write attributes        
        for key, value in self.attributes_serialized.items():
            dataset.attrs[key] = value
        

    def copy(self):
        """ create a copy of the agents """
        return self.__class__.from_state(attributes=self.attributes,
                                         data=self.data)


    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (self.attributes == other.attributes and
                np.array_equal(self.data, other.data))
        
    
    @property
    def droplet_count(self) -> int:
        """ int: the number of agents in this state """
        raise NotImplementedError
    
        
    @property
    def total_amount(self) -> float:
        """ float: total amount of material contained in the agents """
        raise NotImplementedError


    def plot(self, ax=None, *args, **kwargs):
        """ plot the agents """
        pass



class AgentsBase(Parameterized, metaclass=ABCMeta):
    """ represents the dynamics of many agents of the same type """
    
    
    state_class: Any = AgentsStateBase


    def __init__(self, parameters: Dict[str, Any] = None):
        """
        Args:
            parameters (dict):
                Additional parameters. Call
                :meth:`~AgentsBase.show_parameters` for details.
        """        
        super().__init__(parameters)
        self._cache: Dict[str, Any] = {}
        self._logger = logging.getLogger(self.__class__.__name__)
    

    def __repr__(self):
        return (f'{self.__class__.__name__}(parameters={self.parameters!r})')


    def __str__(self):
        return (f'{self.__class__.__name__}(parameters={self.parameters})')
    
    
    @property
    def info(self) -> Dict[str, Any]:
        """ dict: information about the agents """
        return {'class': self.__class__.__name__,
                'parameters': self.parameters.copy()}


    def get_state(self, *args, **kwargs):
        """ create a state associated with this agent """
        if len(args) == 1 and len(kwargs) == 0:
            # exactly a single argument is given
            if isinstance(args[0], self.state_class):
                # a correct instance is given
                return args[0]
            elif isinstance(args[0], dict):
                # instance data is given as a single dictionary
                return self.state_class(**args[0])
        # initialization data is given directly
        return self.state_class(*args, **kwargs)


    def estimate_dt(self, agents_state, background_state) -> float:
        """ estimate the maximal time step for simulating this agent type 
        
        Args:
            agents_state (:class:`AgentsStateBase`):
                The state corresponding to this agent type
            background_state \
                   (:class:`~agent_based.backgrounds.base.BackgroundStateBase`):
                The state corresponding to the background

        Returns:
            float: the maximal time step
        """
        raise NotImplementedError
    
    
    def _check_cache(self, agents_state,
                     background_state: BackgroundStateBase) -> None:
        """ checks whether the simulation needs to run :meth:`_update_cache`.
        
        Subclasses can defined `_update_cache` to populate `self._cache` with
        pre-computed data, which is then available in later.
        
        Args:
            agents_state (:class:`AgentsStateBase`):
                The state of all agents described by this instance
            background_state \
                   (:class:`~agent_based.backgrounds.base.BackgroundStateBase`):
                The state corresponding to the background
        """
        if hasattr(self, '_update_cache'):
            # the class uses a cache internally
            agents_attributes = agents_state.attributes
            if not objects_equal(self._cache.get('agents_attributes'),
                                 agents_attributes):
                # the cache is out-of-date
                self._update_cache(agents_state,  # type: ignore
                                   background_state)
                self._cache['agents_attributes'] = agents_attributes


    def make_evolver_numba(self, agents_state, background_state) -> Callable:
        """ return a function evolve the agents state from time `t` to `t + dt`
        
        Args:
            agents_state (:class:`AgentsStateBase`):
                The state of all agents described by this instance
            background_state (:class:`BackgroundStateBase`):
                The state of the background        

        Returns:
            callable: A function with signature
                (agents_data: :class:`~numpy.ndarray`, t: float, dt: float,
                background_data), evolving `agents_data`
        """
        raise NotImplementedError


    @abstractmethod
    def evolve(self, agents_state, background_state, t: float, dt: float):
        """ evolve the agents state from time `t` to `t + dt`
        
        Args:
            agents_state (:class:`AgentsStateBase`):
                The state of all agents described by this instance
            background_state (:class:`BackgroundStateBase`):
                The state of the background        
            t (float):
                The current time point
            dt (float):
                The time step

        Returns:
            callable: A function with signature
                (background_data: :class:`~numpy.ndarray`, t: float, dt: float),
                which evolves the background_data
        """
        pass



def agents_state_from_hdf(hdf_path) -> AgentsStateBase:
    """ create agents state instance from a stored state
     
    Args:
        hdf_path: HDF path in an already opened file
    """
    if 'class' in hdf_path.attrs:
        # assume everything is stored in root directory
        dataset = hdf_path
    else:
        # assume a single field is stored in the data
        dataset_names = list(hdf_path.keys())
        if len(dataset_names) > 1:
            logging.getLogger(__name__).warning('Found multiple datasets in '
                                                'file. Using only first one.')
            
        dataset = hdf_path[dataset_names[0]]  # retrieve first dataset

    # determine class
    class_name = json.loads(dataset.attrs['class'])
    field_cls = AgentsStateBase._subclasses[class_name]
    
    # load the instance from hdf
    return field_cls.from_hdf_dataset(dataset)



def agents_state_from_file(path: str) -> AgentsStateBase:
    """ create agents state instance from a stored state
     
    Args:
        path (str): Path to the file being read
    """
    import h5py
    
    with h5py.File(path, "r") as fp:
        return agents_state_from_hdf(fp)

