"""
Expands point-like droplet agents to be used with ElasticDroplet

.. autosummary::
   :nosignatures:

   ~HeterogenousAgentsState
   ~HeterogeounsPointAgents

.. codeauthor:: Estefania Vidal <estefania.vidal@ds.mpg.de>
"""

from typing import Dict, Any, Callable, Sequence, Tuple

import numpy as np

from pde.tools import spherical
from pde.tools.numba import jit
from .point_droplet import DropletAgentsState, PointDropletAgents
from ..backgrounds.base import BackgroundStateBase
from .elastic_droplets import ElasticDroplet


class HeterogenousAgentsState(DropletAgentsState):
    """represents the state of droplet agents which are ElasticDroplet"""

    droplet_class = ElasticDroplet


class HeterogenousPointAgents(PointDropletAgents):
    """represents the dynamics of many point droplets in a heterogeneous mesh"""

    def _make_droplet_evolver_numba(
        self, agents_state: DropletAgentsState, background_state: BackgroundStateBase
    ) -> Callable:
        """ create a function to evolve a single agent from time `t` to `t + dt`
        
        Args:
            agents_state \
                (:class:`agent_based.agents.point_droplet.DropletAgentsState`):
                The state of all droplet agents of this instance
            background_state (:class:`BackgroundStateBase`):
                The state of the background

        Returns:
            callable: A function with signature
                (droplet_data: :class:`~numpy.ndarray`, droplet_id: int,
                t: float, dt: float, filed_data: :class:`~numpy.ndarray`,
                field_update: :class:`~numpy.ndarray`), evolving `droplet_data`
                and updating `field_update`
        """
        cEqOut = self._cache["cEqOut"]
        if hasattr(cEqOut, "get_compiled"):
            calc_cEqOut = cEqOut.get_compiled()
        else:
            # try compiling in case cEqOut is a function
            calc_cEqOut = jit(cEqOut)
        cBaseIn = agents_state.parameters["droplet_concentration"]
        #         sBaseIn = self.pde._reaction(c=cBaseIn, t=0)

        radius = spherical.make_radius_from_volume_compiled(self._cache["dim"])
        volume = spherical.make_volume_from_radius_compiled(self._cache["dim"])

        get_concentration = background_state.make_get_concentration_compiled()
        add_amount = background_state.make_add_amount_compiled()

        calc_flux = jit(self._make_flux_outside())

        @jit(nogil=True)
        def droplet_update(
            droplet_data: np.ndarray,
            droplet_id: int,
            background_data,
            t: float,
            dt: float,
        ):
            """update a single droplet based on the surrounding field"""
            R = droplet_data.radius
            V = volume(R)

            # obtain the material flux across the droplet surface
            cInf = get_concentration(background_data, droplet_data.position)
            cEqIn = cBaseIn
            cEqOut = calc_cEqOut(
                droplet_data.position,
                droplet_data.radius,
                droplet_id,
                droplet_data.mesh_size,
                droplet_data.barrier_height,
                t,
            )

            # Calculate the integrated fluxes at the droplet surface. The sign
            # of the fluxes is such that positive values indicate outward fluxes
            flux_out = calc_flux(R, cInf, cEqOut)

            # amount taken up from the outside per sector
            amount_out = -dt * flux_out

            # amount produced in the inside
            sBaseIn = 0
            amount_in = dt * sBaseIn * V

            # update the droplet volume
            dV = (amount_in + amount_out) / cEqIn
            if V + dV < 0:
                # droplet disappears
                amount_out = V * cEqIn - amount_in
                droplet_data.radius = 0.0  # remove all droplet material
            else:
                droplet_data.radius = radius(V + dV)

            # update the scalar field at the droplet surface
            add_amount(background_data, droplet_data.position, -amount_out)

        return droplet_update  # type: ignore
