import numpy as np
from pde.tools.numba import jit
from typing import Dict, Any


class SingleQuench:
    """class calculating equilibrium concentrations"""

    def __init__(
        self,
        q: float,
        c_initial: float = 0.06939,
        Ym: float = 186,
        cc: float = 0.8333,
        t_adim: float = 0.02,
        lambda_break: float = 10,
    ):

        self.cc = cc
        self.Ym = Ym
        self.c_initial = c_initial
        self.q = q
        self.t_adim = t_adim
        self.lambda_break = lambda_break

    def __call__(self, position, radius, droplet_id, mesh_size, c_cav, t):
        """return equilibrium concentration for a position"""
        q = self.q
        t_adim = self.t_adim
        cc = self.cc
        c_initial = self.c_initial
        Ym = self.Ym
        lambda_break = self.lambda_break

        aux = q * t * t_adim
        if aux < 17:
            T = 40 - aux
        else:
            T = 23
        co = (0.0932 * T + 3.211) / 100

        if radius > lambda_break * mesh_size:
            P = cc
        else:
            P = c_cav / (lambda_break - 1) * (radius / mesh_size - 1)

        base_value = co * np.exp(Ym * P / 11000)
        if base_value > c_initial:
            return c_initial
        else:
            return base_value

    def get_compiled(self):
        """create a function calculating equilibrium concentration"""
        q = self.q
        t_adim = self.t_adim
        cc = self.cc
        c_initial = self.c_initial
        Ym = self.Ym
        lambda_break = self.lambda_break

        @jit
        def calc_ceq(position, radius, droplet_id, mesh_size, c_cav, t):
            aux = q * t * t_adim
            if aux < 17:
                T = 40 - aux
            else:
                T = 23
            co = (0.0932 * T + 3.211) / 100

            if radius > lambda_break * mesh_size:
                P = cc
            elif radius < mesh_size:
                P = 0
            else:
                P = c_cav / (lambda_break - 1) * (radius / mesh_size - 1)

            base_value = co * np.exp(Ym * P / 11000)
            if base_value > c_initial:
                return c_initial
            else:
                return base_value

        return calc_ceq

    @property
    def info(self) -> Dict[str, Any]:
        """dict: information about the state"""
        return {
            "type": "TimeDecreasing",
            "q": self.q,
            "lambda_break": self.lambda_break,
            "Young Modulus": self.Ym,
            "P_C": self.cc,
            "t_adim": self.t_adim,
            "initial_concentration": self.c_initial,
        }


class DoubleQuench:
    """class calculating equilibrium concentrations"""

    def __init__(
        self,
        q: float,
        factor: float = 4.0,
        c_initial: float = 0.06939,
        Ym: float = 186,
        cc: float = 0.8333,
        t_adim: float = 0.02,
        lambda_break: float = 10,
        t0=960,
    ):

        self.cc = cc
        self.factor = factor
        self.Ym = Ym
        self.c_initial = c_initial
        self.q = q
        self.t_adim = t_adim
        self.lambda_break = lambda_break
        self.t0 = t0

    def __call__(self, position, radius, droplet_id, mesh_size, c_cav, t):
        """return equilibrium concentration for a position"""
        q = self.q
        factor = self.factor
        t_adim = self.t_adim
        cc = self.cc
        c_initial = self.c_initial
        Ym = self.Ym
        lambda_break = self.lambda_break
        t0 = self.t0

        if t < t0 / t_adim:
            aux = q * t * t_adim
        else:
            aux = q * t0 + factor * q * (t * t_adim - t0)

        if aux < 17:
            T = 40 - aux
        else:
            T = 23
        co = (0.0932 * T + 3.211) / 100

        if radius > lambda_break * mesh_size:
            P = cc
        else:
            P = c_cav / (lambda_break - 1) * (radius / mesh_size - 1)

        base_value = co * np.exp(Ym * P / 11000)
        if base_value > c_initial:
            return c_initial
        else:
            return base_value

    def get_compiled(self):
        """create a function calculating equilibrium concentration"""
        q = self.q
        factor = self.factor
        t_adim = self.t_adim
        cc = self.cc
        c_initial = self.c_initial
        Ym = self.Ym
        lambda_break = self.lambda_break
        t0 = self.t0

        @jit
        def calc_ceq(position, radius, droplet_id, mesh_size, c_cav, t):
            if t < t0 / t_adim:
                aux = q * t * t_adim
            else:
                aux = q * t0 + factor * q * (t * t_adim - t0)

            if aux < 17:
                T = 40 - aux
            else:
                T = 23
            co = (0.0932 * T + 3.211) / 100

            if radius > lambda_break * mesh_size:
                P = cc
            elif radius < mesh_size:
                P = 0
            else:
                P = c_cav / (lambda_break - 1) * (radius / mesh_size - 1)

            base_value = co * np.exp(Ym * P / 11000)
            if base_value > c_initial:
                return c_initial
            else:
                return base_value

        return calc_ceq

    @property
    def info(self) -> Dict[str, Any]:
        """dict: information about the state"""
        return {
            "type": "DoubleQuench",
            "q": self.q,
            "factor": self.factor,
            "lambda_break": self.lambda_break,
            "Young Modulus": self.Ym,
            "P_C": self.cc,
            "t_adim": self.t_adim,
            "initial_concentration": self.c_initial,
            "Change rate time": self.t0,
        }


class NHTimeDecreasing:
    """class calculating equilibrium concentrations"""

    def __init__(
        self,
        q: float,
        c_initial: float = 0.06939,
        Ym: float = 186,
        t_adim: float = 0.02,
    ):

        self.Ym = Ym
        self.c_initial = c_initial
        self.q = q
        self.t_adim = t_adim

    def __call__(self, position, radius, droplet_id, mesh_size, c_cav, t):
        """return equilibrium concentration for a position"""
        q = self.q
        t_adim = self.t_adim
        c_initial = self.c_initial
        Ym = self.Ym

        aux = q * t * t_adim
        if aux < 17:
            T = 40 - aux
        else:
            T = 23
        co = (0.0932 * T + 3.211) / 100

        if radius < mesh_size:
            P = 0
        else:
            stretch = radius / mesh_size
            P = 5 / 6 - 2 / (3 * stretch) - 1 / (6 * stretch ** 4)

        base_value = co * np.exp(Ym * P / 11000)
        if base_value > c_initial:
            return c_initial
        else:
            return base_value

    def get_compiled(self):
        """create a function calculating equilibrium concentration"""
        q = self.q
        t_adim = self.t_adim
        c_initial = self.c_initial
        Ym = self.Ym

        @jit
        def calc_ceq(position, radius, droplet_id, mesh_size, c_cav, t):
            aux = q * t * t_adim
            if aux < 17:
                T = 40 - aux
            else:
                T = 23
            co = (0.0932 * T + 3.211) / 100

            if radius < mesh_size:
                P = 0
            else:
                stretch = radius / mesh_size
                P = 5 / 6 - 2 / (3 * stretch) - 1 / (6 * stretch ** 4)

            base_value = co * np.exp(Ym * P / 11000)
            if base_value > c_initial:
                return c_initial
            else:
                return base_value

        return calc_ceq

    @property
    def info(self) -> Dict[str, Any]:
        """dict: information about the state"""
        return {
            "type": "NeoHookeanTimeDecreasing",
            "q": self.q,
            "Young Modulus": self.Ym,
            "t_adim": self.t_adim,
            "initial_concentration": self.c_initial,
        }
