'''
This module provides the abstract base class for the backgrounds

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import json
from abc import abstractmethod, abstractproperty, ABCMeta
from typing import (Dict, Any, Sequence, Tuple, Callable, Type,  # @UnusedImport
                    TYPE_CHECKING)

import numpy as np

from pde.grids import CartesianGrid
from pde.grids.cartesian import CartesianGridBase
from pde.tools.cuboid import Cuboid
from pde.tools.parameters import Parameterized


if TYPE_CHECKING:
    from ..agents.base import AgentsStateBase  # @UnusedImport



class BackgroundStateBase(metaclass=ABCMeta):
    """ base class for the background of the agent-based simulation """

    _subclasses: Dict[str, Any] = {}  # all classes inheriting from this

    data: np.ndarray


    def __init__(self, data=None, **kwargs):
        """
        Args:
            data (:class:`~numpy.ndarray` or float):
                The data of the background field
        """
        pass
    

    def set_bounds(self, bounds: Sequence[Tuple[float, float]]) -> None:
        """ set the boundaries of the background
        
        Args:
            bounds (sequence):
                A sequence of tuples specifying the lower and upper bound for
                each axis. The number of entries sets the space dimension.
        """
        self._cuboid = Cuboid.from_bounds(np.array(bounds, np.double),
                                          mutable=False)
        self.dim: int = self._cuboid.dim
        self.bounds = self._cuboid.bounds
        self.volume = float(self._cuboid.volume)


    @property
    def grid(self) -> CartesianGridBase:
        """ :class:`pde.grids.cartesian.CartesianGrid`: background grid """
        return CartesianGrid(self.bounds, 1)
        
        
    def __init_subclass__(cls, **kwargs):  # @NoSelf
        """ register all subclassess to reconstruct them later """
        super().__init_subclass__(**kwargs)
        cls._subclasses[cls.__name__] = cls


    @classmethod
    def from_serialized(cls, attributes: Dict[str, str], data: np.ndarray) \
            -> "BackgroundStateBase":
        """ create the background state from serialized data
        
        Args:
            attributes (dict):
                Serialized attributes of the background state
            data (:class:`~numpy.ndarray`):
                The numerical data associated with the background
        """
        # unserialize the attributes
        attributes = cls.unserialize_attributes(attributes)
                
        if 'class' in attributes:
            class_name = attributes.pop('class')
            assert class_name == cls.__name__
            
        # create the instance from the attributes
        return cls(data=data, **attributes)

   
    @classmethod
    def from_hdf_dataset(cls, dataset) -> "BackgroundStateBase":
        """ construct the background by reading data from an hdf5 dataset """
        # copy attributes from hdf
        attributes = dict(dataset.attrs)
        
        # determine class
        class_name = json.loads(attributes.pop('class'))
        field_cls = cls._subclasses[class_name]
        
        # construct the instance
        obj = field_cls.from_serialized(attributes, data=dataset)
        return obj  # type: ignore

    
    @property
    def attributes(self) -> Dict[str, Any]:
        """ dict: information about the background """
        return {'class': self.__class__.__name__,
                'bounds': self.bounds}


    @property
    def attributes_serialized(self) -> Dict[str, str]:
        """ dict: serialized version of the attributes """
        return {key: json.dumps(value)
                for key, value in self.attributes.items()}
    

    @classmethod
    def unserialize_attributes(cls, attributes: Dict[str, str]) \
            -> Dict[str, Any]:
        """ unserializes the given attributes
        
        Args:
            attributes (dict):
                The serialized attributes
                
        Returns:
            dict: The unserialized attributes
        """
        return {key: json.loads(value)
                for key, value in attributes.items()}
    

    def to_file(self, filename: str, **kwargs):
        r""" store background in a file
        
        Args:
            filename (str):
                Path where the data is stored
            \**kwargs:
                Additional parameters may be supported for some formats 
        """
        import h5py
        with h5py.File(filename, "w") as fp:
            self._write_hdf_dataset(fp, **kwargs)
                    
    
    def _write_hdf_dataset(self, hdf_path, key: str = 'data'):
        """ write data to a given hdf5 file pointer `hdf_path` """
        dataset = hdf_path.create_dataset(key, data=self.data)
        
        # write attributes        
        for key, value in self.attributes_serialized.items():
            dataset.attrs[key] = value


    @abstractmethod
    def copy(self): pass


    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (self.attributes == other.attributes and
                np.array_equal(self.data, other.data))
        

    def plot(self, ax=None, **kwargs):
        """ plot the background field """
        pass


    @abstractproperty
    def total_amount(self) -> float:
        """ float: the total material amount in the background """
        pass
    

    @abstractmethod
    def get_concentration(self, points):
        """ determine concentration at the given points
        
        Args:
            points (:class:`~numpy.ndarray`):
                The coordinates of the single point or the list of points at
                which the concentration is returned
        """
        pass
    

    @abstractmethod
    def add_amount(self, point: np.ndarray, amount: float):
        """ add the given amount to the background
        
        Args:
            point (:class:`~numpy.ndarray`):
                Point where the amount is added to the background
            amount (float):
                The total amount added to the background
        """
        pass

    
    def make_get_concentration_compiled(self) -> Callable:
        """ get a compiled function for obtaining concentrations
        
        Returns:
            callable: a function with signature (data: :class:`~numpy.ndarray`,
            point: :class:`~numpy.ndarray`), which determines the concentration
            at point `point` given the background state `data`.
        """
        raise NotImplementedError

    
    def make_add_amount_compiled(self) -> Callable:
        """ get a compiled function for adding amount to the background
        
        Returns:
            callable: a function with signature (data: :class:`~numpy.ndarray`,
            point: :class:`~numpy.ndarray`, amount: float), which adds `amount`
            to the background state given by `data` at point `point`.
        """
        raise NotImplementedError



class BackgroundBase(Parameterized, metaclass=ABCMeta):
    """ base class for the specifying the background dynamics """ 
    
    state_class = BackgroundStateBase  # class describing the associated state


    def __repr__(self):
        return (f'{self.__class__.__name__}(parameters={self.parameters!r})')


    def __str__(self):
        return (f'{self.__class__.__name__}(parameters={self.parameters})')


    @property
    def info(self) -> Dict[str, Any]:
        """ dict: information about the background """
        return {'class': self.__class__.__name__,
                'parameters': self.parameters}


    def get_state(self, *args, **kwargs):
        """ create a state associated with this background """
        return self.state_class(*args, **kwargs)


    @abstractmethod
    def estimate_dt(self, state) -> float:
        """ get the optimal time step for the simulation of the background
        
        Args:
            state (:class:`ScalarBackgroundState`):
                The background state
                
        Returns:
            float: the time step
        """
        pass


    @abstractmethod
    def evolve(self, state,
               t: float, dt: float,
               agents_state: "AgentsStateBase"):
        """ evolve the field from time `t` to `t + dt`
        
        Args:
            state (:class:`MeanfieldBackgroundState`):
                The state of the background
            t (float):
                The current time point
            dt (float):
                The time step used to evolve the state
            agents_state (:class:`agent_based.agents.base.AgentsStateBase`):
                The state of the agents
        """
        pass


    def make_evolver_numba(self, state) -> Callable:
        """ return function evolving background state from time `t` to `t + dt`
        
        Args:
            state (:class:`BackgroundStateBase`):
                The state of the background        

        Returns:
            callable: A function with signature (background_data, t: float,
                dt: float, agents_data), which evolves the background_data.
        """
        raise NotImplementedError



def background_state_from_hdf(hdf_path) -> BackgroundStateBase:
    """ create background state instance from a stored state
     
    Args:
        hdf_path: HDF Path in an already opened file
    """
    # a single field is stored in the data
    dataset = hdf_path[list(hdf_path.keys())[0]]  # retrieve only dataset
    
    # determine class
    class_name = json.loads(dataset.attrs['class'])
    field_cls = BackgroundStateBase._subclasses[class_name]
    
    # load the instance from hdf
    return field_cls.from_hdf_dataset(dataset)  # type: ignore



def background_state_from_file(path: str) -> BackgroundStateBase:
    """ create background state instance from a stored state
     
    Args:
        path (str): Path to the file being read
    """
    import h5py
    
    with h5py.File(path, "r") as fp:
        return background_state_from_hdf(fp)

