"""
Defines backgrounds represented by a scalar field, i.e., with locally resolved
values.

.. autosummary::
   :nosignatures:

   ~ScalarBackgroundState
   ~ScalarPDEBackground
   ~DiffusionBackground

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

import json
from abc import ABCMeta, abstractmethod
from typing import Dict, Any, Callable, Type, Optional, TYPE_CHECKING  # @UnusedImport

import numpy as np

from pde.grids.cartesian import GridBase, CartesianGridBase
from pde.fields import ScalarField
from pde.pdes.base import PDEBase
from pde.tools.numba import jit
from pde.tools.docstrings import get_text_block
from pde.tools.parameters import Parameter
from pde.tools.typing import ArrayLike

from .base import BackgroundStateBase, BackgroundBase


if TYPE_CHECKING:
    from ..agents.base import AgentsStateBase  # @UnusedImport


class ScalarBackgroundState(BackgroundStateBase):
    """the state associated with a spatially resolved background"""

    def __init__(self, grid: GridBase, data: ArrayLike = None, label: str = None):
        """
        Args:
            grid (:class:`~pde.grids.GridBase`):
                Grid defining the space on which this field is defined
            data (:class:`~numpy.ndarray` or float, optional):
                Field values at the support points of the grid
            label (str, optional):
                Name of the field
        """
        super().__init__(self)

        if not isinstance(grid, CartesianGridBase):
            raise NotImplementedError(
                "Agent simulation has only been "
                "implemented for Cartesian grids and not "
                f"for {grid.__class__.__name__}"
            )

        self._field = ScalarField(grid, data, label=label)
        self.data = self._field.data
        self.set_bounds(grid.axes_bounds)

    @classmethod
    def from_field(cls, field: ScalarField):
        """create a scalar background state from a scalar field

        Args:
            field (:class:`~pde.fields.scalar.ScalarField`):
                The scalar field that initializes the background

        Returns:
            :class:`ScalarBackgroundState`: The initialized instance
        """
        return cls(field.grid, field.data, field.label)

    def copy(self) -> "ScalarBackgroundState":
        return self.__class__(grid=self.grid.copy(), data=self.data, label=self.label)

    @property
    def attributes(self) -> Dict[str, Any]:
        """dict: information about the background"""
        result = self._field.attributes
        result["class"] = self.__class__.__name__
        return result

    @property
    def attributes_serialized(self) -> Dict[str, str]:
        """dict: serialized version of the attributes"""
        results = {}
        for key, value in self.attributes.items():
            if key == "grid":
                results[key] = value.state_serialized
            else:
                results[key] = json.dumps(value)
        return results

    @property
    def grid(self) -> CartesianGridBase:
        """:class:`~pde.grids.cartesian.CartesianGridBase`: the grid"""
        return self._field.grid  # type: ignore

    @property
    def field(self) -> ScalarField:
        """:class:`~pde.fields.scalar.ScalarField`: the scalar field"""
        return self._field

    @property
    def label(self) -> Optional[str]:
        return self._field.label

    @label.setter
    def label(self, value: Optional[str]) -> None:
        self._field.label = value

    def plot(self, ax=None, **kwargs):
        """plot the background as a scalar field

        This simply calls :meth:`~pde.fields.base.DataFieldBase.plot` and all
        arguments are forwarded.
        """
        return self._field.plot(ax=ax, **kwargs)

    @classmethod
    def from_hdf_dataset(cls, dataset) -> "ScalarBackgroundState":
        """construct the state by reading data from an hdf5 dataset"""
        # copy attributes from hdf
        attributes = dict(dataset.attrs)
        attributes.pop("class", None)  # remove 'class' if it is present

        # unserialize the attributes
        attributes = ScalarField.unserialize_attributes(attributes)
        field = ScalarField.from_state(attributes, data=dataset)
        return cls.from_field(field)  # type: ignore

    @property
    def total_amount(self) -> float:
        """float: the total material amount in the background"""
        return self._field.integral.real

    def get_concentration(self, points: np.ndarray):
        """determine concentration at the given points

        Args:
            points (:class:`~numpy.ndarray`):
                The coordinates of the single point or the list of points at
                which the concentration is returned
        """
        return self._field.interpolate(points)

    def add_amount(self, point: np.ndarray, amount: float):
        """add the given amount to the background

        Args:
            point (:class:`~numpy.ndarray`):
                Point where the amount is added to the background
            amount (float):
                The total amount added to the background
        """
        self._field.add_interpolated(point, amount)

    def make_get_concentration_compiled(self) -> Callable:
        """get a compiled function for obtaining concentrations

        Returns:
            callable: a function with signature (data: :class:`~numpy.ndarray`,
            point: :class:`~numpy.ndarray`), which determines the concentration
            at point `point` given the background state `data`.
        """
        return self._field.grid.make_interpolator_compiled()

    def make_add_amount_compiled(self) -> Callable:
        """get a compiled function for adding amount to the background

        Returns:
            callable: a function with signature (data: :class:`~numpy.ndarray`,
            point: :class:`~numpy.ndarray`, amount: float), which adds `amount`
            to the background state given by `data` at point `point`.
        """
        return self._field.grid.make_add_interpolated_compiled()


class ScalarBackgroundBase(BackgroundBase, metaclass=ABCMeta):
    """base class for a background based on a scalar field"""

    state_class: Type[ScalarBackgroundState] = ScalarBackgroundState

    def get_state(
        self, grid_or_field, data: ArrayLike = None  # type: ignore
    ) -> ScalarBackgroundState:
        """ create a state associated with this background
        
        Args:
            grid_or_field (:class:`~pde.grids.cartesian.CartesianGridBase` or \
                    :class:`~pde.fields.scalar.ScalarField`):
                The grid or the field that defines this background
            data (:class:`~numpy.ndarray` or float):
                The initial concentration field of the background
        """
        if isinstance(grid_or_field, CartesianGridBase):
            grid = grid_or_field
            data = None
        elif isinstance(grid_or_field, ScalarField):
            grid = grid_or_field.grid  # type: ignore
            data = grid_or_field.data if data is None else data
        else:
            raise TypeError(
                f"The state needs to be initialized with a scalar "
                "field or a Cartesian grid"
            )
        return self.state_class(grid, data)

    def estimate_dt(self, state: ScalarBackgroundState) -> float:
        """get the optimal time step for the simulation of the background

        Args:
            state (:class:`ScalarBackgroundState`):
                The background state

        Returns:
            float: the time step
        """
        raise NotImplementedError

    def make_evolver_numba(self, state: ScalarBackgroundState) -> Callable:
        """return a function evolve the field from time `t` to `t + dt`

        Args:
            state (:class:`ScalarBackgroundState`):
                The background state

        Returns:
            callable: A function with signature (background_data, t: float,
                dt: float, agents_data), which evolves the background_data.
        """
        raise NotImplementedError

    @abstractmethod
    def evolve(
        self,
        state: ScalarBackgroundState,
        t: float,
        dt: float,
        agents_state: "AgentsStateBase" = None,
    ):
        """evolve the field from time `t` to `t + dt`

        Args:
            state (:class:`MeanfieldBackgroundState`):
                The state of the background
            t (float):
                The current time point
            dt (float):
                The time step used to evolve the state
            agents_state (:class:`agent_based.agents.base.AgentsStateBase`):
                The state of the agents (Not used by this class)
        """
        pass


class ScalarPDEBackground(ScalarBackgroundBase):
    """background based on a scalar field evolving according to a PDE"""

    def __init__(self, pde: PDEBase, parameters: Dict[str, Any] = None):
        """initialize the scalar background field and its PDE

        Args:
            pde (:class:`~pde.pdes.base.PDEBase`):
                The partial differential equation describing the dynamics of the
                scalar background field.
            parameters (dict):
                Additional parameters. Call
                :meth:`~ScalarPDEBackground.show_parameters` for details.
        """
        super().__init__(parameters=parameters)
        self.pde = pde

    @property
    def info(self) -> Dict[str, Any]:
        """dict: information about the background"""
        result = super().info
        result["pde"] = {"class": self.pde.__class__.__name__}
        return result

    def make_evolver_numba(self, state: ScalarBackgroundState) -> Callable:
        """return a function evolve the field from time `t` to `t + dt`

        Args:
            state (:class:`ScalarBackgroundState`):
                The background state

        Returns:
            callable: A function with signature (background_data, t: float,
                dt: float, agents_data), which evolves the background_data.
        """
        pde_rhs = self.pde._make_pde_rhs_numba(state._field)

        @jit
        def evolver(background_data, t: float, dt: float, agents_data):
            """evolve the diffusion equation explicitly"""
            background_data += dt * pde_rhs(background_data, t)

        return evolver  # type: ignore

    def evolve(
        self,
        state: ScalarBackgroundState,
        t: float,
        dt: float,
        agents_state: "AgentsStateBase" = None,
    ):
        """evolve the field from time `t` to `t + dt`

        Args:
            state (:class:`MeanfieldBackgroundState`):
                The state of the background
            t (float):
                The current time point
            dt (float):
                The time step used to evolve the state
            agents_state (:class:`agent_based.agents.base.AgentsStateBase`):
                The state of the agents (Not used by this class)
        """
        rate = self.pde.evolution_rate(state._field, t)
        state._field += dt * rate


class DiffusionBackground(ScalarPDEBackground):
    """background based on a scalar field evolving with simple diffusion"""

    parameters_default = [
        Parameter(
            "diffusivity",
            1,
            float,
            "Diffusivity in the background field. This class only "
            "supports constant diffusivities. Diffusivities depending "
            "on local concentration are supported by the "
            "ReactionDiffusionPDE class.",
        ),
        Parameter(
            "boundary_conditions",
            "natural",
            object,
            "Defines the boundary conditions on the background field."
            + get_text_block("ARG_BOUNDARIES"),
        ),
    ]

    def __init__(self, parameters: Dict[str, Any] = None):
        """initialize the background

        Args:
            parameters (dict):
                Additional parameters. Call
                :meth:`~DiffusionBackground.show_parameters` for details.
        """
        from pde import DiffusionPDE

        # skip calling the parent init since it expects the pde, but we first
        # need to parse the parameters. We thus simply call the grand-parent
        # init method directly
        ScalarBackgroundBase.__init__(self, parameters=parameters)

        # initialize diffusion equation
        self.pde = DiffusionPDE(
            diffusivity=self.parameters["diffusivity"],
            bc=self.parameters["boundary_conditions"],
        )

    def estimate_dt(self, state: ScalarBackgroundState) -> float:
        """get the optimal time step for the simulation of the background

        Returns:
            float: the time step
        """
        dx = float(state.grid.discretization.min())
        return 0.1 * dx ** 2 / float(self.pde.diffusivity)  # type: ignore
