"""
Provides a class representing the full system state.

.. autosummary::
   :nosignatures:

   ~SimulationState
   ~AgentSimulation

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

import logging
import inspect
from typing import Callable, Union, Sequence, Tuple, Any, Dict

import numba as nb

from pde.solvers.controller import Controller, TRangeType, TrackerCollectionDataType
from pde.tools.numba import jit
from pde.tools.misc import hdf_write_attributes
from pde.tools.plotting import plot_on_axes

from .backgrounds.base import (
    BackgroundStateBase,
    BackgroundBase,
    background_state_from_hdf,
)
from .agents.base import AgentsStateBase, AgentsBase, agents_state_from_hdf


class SimulationState:
    """Class defining the state"""

    def __init__(self, background: BackgroundStateBase, agents: AgentsStateBase = None):
        """
        Args:
            background
                   (:class:`~agent_based.backgrounds.base.BackgroundStateBase`):
                The instance describing the background
            agents (:class:`~agent_based.agents.base.AgentsStateBase`):
                The instance describing all the agents. If omitted, no agents
                are added.
        """
        # initialize the background
        if not isinstance(background, BackgroundStateBase):
            raise TypeError(
                f"The first argument to {self.__class__.__name__} "
                "must be of type BackgroundStateBase and not "
                f"{background.__class__.__name__}"
            )
        self.background = background
        self.agents = agents

    @classmethod
    def from_hdf_dataset(cls, dataset) -> "SimulationState":
        """construct the instance by reading data from an hdf5 dataset

        Args:
            dataset: the hdf5 dataset (in an already opened file)
        """
        background = background_state_from_hdf(dataset["background"])
        agents = agents_state_from_hdf(dataset["agents"])
        return cls(background, agents)

    @classmethod
    def from_file(cls, path: str) -> "SimulationState":
        """create simulation state instance from data stored in a hdf file

        Args:
            path (str): Path to the hdf file being read
        """
        import h5py

        with h5py.File(path, "r") as fp:
            return cls.from_hdf_dataset(fp)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(background={self.background!r}, "
            f"agents={self.agents!r})"
        )

    def __str__(self):
        return (
            f"{self.__class__.__name__}(background={self.background!s}, "
            f"agents={self.agents!s})"
        )

    def copy(self) -> "SimulationState":
        """copy the state"""
        return self.__class__(
            background=self.background.copy(), agents=self.agents.copy()
        )

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.background == other.background and self.agents == other.agents

    @property
    def attributes(self) -> Dict[str, Any]:
        """dict: information about the state"""
        return {
            "background": self.background.attributes,
            "agents": self.agents.attributes,
        }

    def _write_hdf_dataset(self, hdf_path):
        """write data to a given hdf5 file pointer `hdf_path`"""
        self.background._write_hdf_dataset(hdf_path.create_group("background"))
        self.agents._write_hdf_dataset(hdf_path.create_group("agents"))

    def to_file(self, filename: str, info: Dict[str, Any] = None) -> None:
        r"""store agents state in a hdf file

        Args:
            filename (str):
                Path where the data is stored
            info (dict):
                Extra information that is written to the hdf attributes. Note
                that the values in this dictionary will be JSON-serialized.
        """
        import h5py

        with h5py.File(filename, "w") as fp:
            self._write_hdf_dataset(fp)
            hdf_write_attributes(fp, info)

    @property
    def data(self) -> Tuple[Any, Any]:
        """the full data of the simulation"""
        return self.background.data, self.agents.data

    @property
    def droplet_count(self) -> int:
        """int: the number of droplets in the full simulation

        This only counts droplets with non-zero radius.
        """
        return self.agents.droplet_count

    @property
    def total_amount(self) -> float:
        """float: total amount of material in the state"""
        return self.background.total_amount + self.agents.total_amount

    @plot_on_axes()
    def plot(
        self,
        ax,
        elements: str = "all",
        background_args: Dict[str, Any] = None,
        agent_args: Dict[str, Any] = None,
    ):
        r"""plot the emulsion together with the background field

        Args:
            ax (:class:`matplotlib.axes.Axes`):
                The axes in which the simulation state is shown.
            elements (str):
                Determines which elements are plotted. Possible values are
                `all`, `droplets`, or `background`.
            background_args (dict):
                Additional arguments for the background plot
            agent_args (dict):
                Additional arguments for the agents plot
        """
        if elements not in {"all", "agents", "droplets", "background"}:
            raise ValueError("`elements` must be `all`, `agents`, or " "`background`")

        if background_args is None:
            background_args = {}
        if agent_args is None:
            agent_args = {}

        if elements in {"all", "background"}:
            # plot the background
            background_args.setdefault("colorbar", True)
            self.background.plot(ax=ax, **background_args)

        if elements in {"all", "agents", "droplets"}:
            # plot the agents
            agent_args.setdefault("grid", self.background.grid)
            agent_args.setdefault("color", "w" if elements == "all" else "k")
            agent_args.setdefault("fill", True)
            self.agents.plot(ax=ax, **agent_args)

        # set dimensions
        bounds = self.background.bounds
        ax.set_xlim(*bounds[0])
        ax.set_ylim(*bounds[1])


class AgentSimulation:
    """Class defining the agent-based simulation"""

    def __init__(
        self,
        background: BackgroundBase,
        agents: Union[AgentsBase, Sequence[AgentsBase]] = None,
    ):
        """
        Args:
            background \
                   (:class:`~agent_based.backgrounds.base.BackgroundStateBase`):
                The instance describing the background
            agents (:class:`~agent_based.agents.base.AgentsBase`):
                The instance describing all the agents. If omitted, no agents
                are added.
        """
        # initialize the background
        if inspect.isclass(background):
            background = background()  # type: ignore

        if not isinstance(background, BackgroundBase):
            raise TypeError(
                f"The first argument to {self.__class__.__name__} "
                "must be of type BackgroundBase"
            )
        self.background = background
        self.agents = agents
        self._logger = logging.getLogger(self.__class__.__name__)

    def __repr__(self):
        """return instance as string"""
        return (
            f"{self.__class__.__name__}(background={self.background!r}, "
            f"agents={self.agents!r})"
        )

    def __str__(self):
        """return instance as string"""
        return (
            f"{self.__class__.__name__}(background={self.background!s}, "
            f"agents={self.agents!s})"
        )

    @property
    def info(self) -> Dict[str, Any]:
        """dict: information about the state"""
        return {"background": self.background.info, "agents": self.agents.info}

    def get_state(self, background, agents=None) -> SimulationState:
        """create a state associated with this agent

        Args:
            background:
                Data describing the background. This can either be an instance
                of :class:`~agent_based.backgrounds.base.BackgroundStateBase` or
                a dictionary with appropriate data for the background class.
            agents:
                Data describing the agents. This can either be an instance
                of :class:`~agent_based.agents.base.AgentsBase` or a dictionary
                with appropriate data for the chosen agent class.
        """
        if isinstance(background, BackgroundStateBase):
            background_state = background.copy()
        elif isinstance(background, dict):
            background_state = self.background.get_state(**background)
        else:
            background_state = self.background.get_state(background)

        if agents is None:
            agents_state = None
        elif isinstance(agents, AgentsStateBase):
            agents_state = agents.copy()
        elif isinstance(agents, dict):
            agents_state = self.agents.get_state(**agents)
        else:
            agents_state = self.agents.get_state(agents)

        return SimulationState(background_state, agents_state)

    def estimate_dt(self, state: SimulationState) -> float:
        """get the optimal time step for the simulation

        Returns:
            float: the time step
        """
        # try obtaining dt from background
        try:
            dts = [self.background.estimate_dt(state.background)]
        except NotImplementedError:
            self._logger.info("Could not determine time step for background")
            dts = []

        # try obtaining dt from agents
        try:
            dts.append(self.agents.estimate_dt(state.agents, state.background))
        except NotImplementedError:
            self._logger.info("Could not determine time step for agents")

        if dts:
            return min(dts)
        else:
            self._logger.warn("Chose fixed default time step of 1e-3")
            return 1e-3  # default value

    def make_evolver_numba(self, state: SimulationState) -> Callable:
        """return a function evolving the state from time `t` to `t + dt`

        Args:
            state (:class:`SimulationState`):
                The state of the simulation

        Returns:
            callable: A function with signature (state_data, t: float,
            dt: float), which evolves the state in time
        """

        # obtain the compiled evolvers of the background and the agents
        bg_evolver = self.background.make_evolver_numba(state.background)
        agent_evolver = self.agents.make_evolver_numba(state.agents, state.background)

        # determine the signature for eager compilation
        signature = nb.void(nb.typeof(state.data), nb.double, nb.double)

        @jit(signature)
        def evolver(state_data, t: float, dt: float):
            """evolve the agents and the background"""
            background_data, agents_data = state_data
            agent_evolver(agents_data, background_data, t, dt)
            bg_evolver(background_data, t, dt, agents_data)

        return evolver  # type: ignore

    def evolve(self, state: SimulationState, t: float, dt: float):
        """evolve the state from time `t` to `t + dt`

        Args:
            state (:class:`SimulationState`):
                The state of the simulation
            t (float):
                The current time point
            dt (float):
                The time step
        """
        self.agents.evolve(state.agents, state.background, t, dt)
        self.background.evolve(state.background, t, dt, state.agents)

    def run(
        self,
        state: SimulationState,
        t_range: TRangeType,
        dt: float = None,
        tracker: TrackerCollectionDataType = ["progress"],
        backend: str = "auto",
    ) -> SimulationState:
        """run an agent-based simulation

        Args:
            state (:class:`SimulationState`):
                The initial state
            t_range (float or tuple of floats):
                Sets the time range for which the simulation is run. If only a
                single value `t_end` is given, the time range is assumed to be
                `[0, t_end]`.
            dt (float):
                Time step of the explicit stepping. If `None`, the time step
                will be chosen automatically using the method
                :func:`~agent_based.state.AgentSimulation.estimate_dt`.
            tracker:
                Defines trackers that process the state of the simulation at
                fixed time intervals. Multiple trackers can be specified as a
                list. The default value simply displays a progress bar. To
                disable trackers, set the value to `None`.
            backend (str):
                Determines how the function is created. Accepted  values are
                'numpy` and 'numba'. Alternatively, 'auto' lets the code decide
                for the most optimal backend.

        Returns:
            :class:`SimulationState`:
                The state of the simulation at the last time point
        """
        from .solver import AgentSolver

        solver = AgentSolver(self, backend=backend)
        controller = Controller(solver, t_range=t_range, tracker=tracker)
        return controller.run(state, dt)  # type: ignore
