r"""
Provides infrastructure to solve agent-based simulations

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

from typing import Callable, Dict, Any  # @UnusedImport
import logging

import numpy as np

from pde.solvers.base import SolverBase

from .simulation import AgentSimulation, SimulationState



class AgentSolver(SolverBase):
    """ Solver for agent-based simulation of emulsions """

    name = 'agents'


    def __init__(self, simulation: AgentSimulation, backend: str = 'auto'):
        """ initialize the explicit solver for the agent-based simulation
        
        Args:
            simulation (:class:`~agent_based.simulation.AgentSimulation`):
                The simulation that will be run. This defines the behavior of
                the background and the agents.
            backend (str):
                Determines how the function is created. Accepted  values are
                'numpy` and 'numba'. Alternatively, 'auto' lets the code decide
                for the most optimal backend.
        """        
        self.info: Dict[str, Any] = {}
        self._logger = logging.getLogger(self.__class__.__name__)
        self.simulation = simulation
        self.backend = backend


    def _make_stepper_numpy(self, dt: float) -> Callable:
        """ return function evolving state from time `t_start` to `t_end`
        
        Args:
            dt (float): The time step
            
        Returns:
            callable: Function with signature (state: SimulationState,
            t_start: float, t_end: float), which advances `state` in time.
        """

        def stepper(state: SimulationState, t_start: float, t_end: float) \
                -> float:
            """ function that advances the state from t_start to t_end """
            steps = max(1, int(np.ceil((t_end - t_start) / dt)))

            for step in range(steps):
                t = t_start + step * dt  # advance time
                self.simulation.evolve(state, t, dt)
                
            self.info['steps'] += steps

            return t + dt
        
        return stepper


    def _make_stepper_numba(self, state: SimulationState, dt: float) \
            -> Callable:
        """ return function evolving state from time `t_start` to `t_end`
        
        This function used compiled evolvers for the background and the agents.
        
        Args:
            state (:class:`~agent_based.simulation.SimulationState`):
                An example of the simulation state
            dt (float):
                The time step
                
        Returns:
            callable: Function with signature (state: SimulationState,
            t_start: float, t_end: float), which advances `state` in time.
        """

        simulation_evolver = self.simulation.make_evolver_numba(state)

        def stepper(state: SimulationState, t_start: float, t_end: float) \
                -> float:
            """ function that advances the state from t_start to t_end """
            steps = max(1, int(np.ceil((t_end - t_start) / dt)))

            for step in range(steps):
                t = t_start + step * dt  # advance time
                simulation_evolver(state.data, t, dt)

            self.info['steps'] += steps

            return t + dt

        return stepper


    def make_stepper(self, state: SimulationState, dt: float = None) \
            -> Callable:
        r""" return a stepper function using an explicit scheme
        
        Note that if the `numba` backend is chosen, the state supplied to this
        function must be the identical state that is also used in the stepper.  
        
        Args:
            state (:class:`~agent_based.simulation.SimulationState`):
                An example of the simulation state, which is used to extract the
                grid and other information.
            dt (float):
                Time step of the explicit stepping. If `None`, the time step
                will be chosen automatically using the method
                :func:`~agent_based.state.AgentSimulation.estimate_dt`.
            \**kwargs: These are currently ignored

        Returns:
            Function that can be called to advance the `state` from time
            `t_start` to time `t_end`. The function call signature is
            `(state: AgentSimulation, t_start: float, t_end: float)`        
        """
#         self.simulation.agents._prepare_evolver(state.agents,state.background)
        
        if dt is None:
            dt = self.simulation.estimate_dt(state)
            if np.isinf(dt):
                # this can happen if there are no restrictions on the time step
                dt = 1e3

        # store information about the simulation
        try:
            self.info['agent_count'] = state.agents.droplet_count
        except NotImplementedError:
            self.info['agent_count'] = 'N/A'
        self.info['dt'] = dt
        self.info['steps'] = 0

        if self.backend == 'auto':
            try:
                return self._make_stepper_numba(state, dt)
            except NotImplementedError:
                self._logger.warning('Numba backend is not implemented for all '
                                     'parts of the simulation.')
                return self._make_stepper_numpy(dt)
                
        elif self.backend == 'numba':
            return self._make_stepper_numba(state, dt)
        elif self.backend == 'numpy':
            return self._make_stepper_numpy(dt)
        else:
            raise ValueError(f'Unknown backend `{self.backend}`')

    