'''
Provides classes that track simulation results in an agent-based simulation

.. autosummary::
   :nosignatures:

   ~BackgroundTracker
   ~DropletAgentTracker

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

from typing import Union
import logging

from pde.fields.base import FieldBase
from pde.trackers.base import TrackerBase, IntervalData, InfoDict
from droplets.droplet_tracks import DropletTrack, DropletTrackList
from droplets.emulsions import EmulsionTimeCourse

from .simulation import SimulationState

        
        
class BackgroundTracker(TrackerBase):
    """ tracker for analyzing the background of an agent-based simulations
    
    This acts as a wrapper around any of the trackers from :mod:`pde.trackers`,
    e.g., `tracker = BackgroundTracker(PlotTracker)`.
    """
    
    def __init__(self, tracker: TrackerBase):
        """
        Args:
            tracker (TrackerBase):
                The tracker that will receive the background field
        """ 
        self.tracker = tracker
        self.interval = tracker.interval
        self._logger = logging.getLogger(self.__class__.__name__)

    
    def initialize(self, state: SimulationState,  # type: ignore
                   info: InfoDict = None) -> float:
        """ initialize the tracker with information about the simulation
        
        Args:
            state (:class:`~agent_based.state.SimulationState`):
                An example of the data that will be analyzed by the tracker
            info (dict):
                Extra information from the simulation        
                
        Returns:
            float: The first time the tracker needs to handle data
        """
        if isinstance(state.background, FieldBase):
            field = state.background
        elif (hasattr(state.background, '_field') and 
                isinstance(state.background._field, FieldBase)):  # type: ignore
            field = state.background._field  # type: ignore
        else:
            self._logger.warning("Background does not seem to contain a scalar "
                                 "field")
        return self.tracker.initialize(field, info)
        
    
    def handle(self, state: SimulationState, t: float) -> None:  # type: ignore
        """ handle data supplied to this tracker
        
        Args:
            field (:class:`~agent_based.state.SimulationState`):
                The current state of the simulation
            t (float): The associated time
        """
        if isinstance(state.background, FieldBase):
            field = state.background
        elif (hasattr(state.background, '_field') and 
                isinstance(state.background._field, FieldBase)):  # type: ignore
            field = state.background._field  # type: ignore
        else:
            self._logger.warning("Background does not seem to contain a scalar "
                                 "field")
        self.tracker.handle(field, t)
    
    
    def finalize(self, info: InfoDict = None) -> None:
        """ finalize the tracker, supplying additional information

        Args:
            info (dict):
                Extra information from the simulation        
        """
        self.tracker.finalize(info)
        
        
        
class DropletAgentTracker(TrackerBase):
    """ Tracker storing information about droplets in an agent based simulation
    
    Attributes:
        emulsions (:class:`~droplets.analysis.emulsions.EmulsionTimeCourse`):
            An object describing the emulsion at the determined intervals
        droplet_tracks (:class:`~droplets.analysis.droplets.DropletTrackList`):
            An object describing the time course of individual droplets.
            
    The two attributes `emulsions` and `droplet_tracks` contain equivalent
    information, but their structure is different and either one might thus be
    used to analyze the agent based simulation.
    """
    
    def __init__(self, interval: IntervalData = 1,
                 store_emulsions: Union[bool, str] = True,
                 store_droplet_tracks: Union[bool, str] = True,
                 keep_vanished: bool = False):
        """
        Args:
            interval
                Determines how often the tracker interrupts the simulation.
                Simple numbers are interpreted as durations measured in the
                simulation time variable. Alternatively, instances of
                :class:`~droplets.simulation.trackers.LogarithmicIntervals` and
                :class:`~droplets.simulation.trackers.RealtimeIntervals`
                might be given for more control.
            store_emulsions (bool or str):
                Determines whether to store data on emulsions in an instance of
                :class:`~droplets.analysis.emulsions.EmulsionTimeCourse`. No
                data is stored when this is `False`. Otherwise, the data is
                available in the :attr:`emulsions` attributed of the tracker
                instance. The data is additionally written to a file when a
                path is supplied as a string.
            store_droplet_tracks (bool or str):
                Determines whether to store data on droplets in an instance of
                :class:`~droplets.analysis.droplets.DropletTrackList`. No
                data is stored when this is `False`. Otherwise, the data is
                available in the :attr:`droplet_tracks` attributed of the
                tracker instance. The data is additionally written to a file
                when a path is supplied as a string.
            keep_vanished (bool):
                Flag determining whether vanished droplets (with zero radius)
                are still stored. The default is to filter these droplets.
        """
        super().__init__(interval=interval)
        self.store_emulsions = store_emulsions
        self.store_droplet_tracks = store_droplet_tracks
        self.keep_vanished = keep_vanished
        

    def initialize(self, state: SimulationState,  # type: ignore
                   info: InfoDict = None) -> float:
        """ 
        Args:
            state (:class:`SimulationState`):
                The initial state of the agent based simulation
            info (dict):
                Extra information from the simulation        
        """
        if not isinstance(state, SimulationState):
            self._logger.warning('state is not of type SimulationState')
        
        # initialize the tracked data
        if self.store_emulsions is not False:
            self.emulsions = EmulsionTimeCourse()
            self.emulsions.grid = state.background.grid  # store the grid, too
            
        if self.store_droplet_tracks is not False:
            tracks = [DropletTrack() for _ in range(len(state.agents.droplets))]
            self.droplet_tracks = DropletTrackList(tracks)
            
        return super().initialize(state.background, info)  # type: ignore
        
        
    def handle(self, state: SimulationState, t: float) -> None:  # type: ignore
        """ handle data supplied to this tracker
        
        Args:
            state (:class:`SimulationState`):
                The current state of the agent based simulation
            t (float): The associated time
        """
        # handle emulsion time course
        if self.store_emulsions is not False:
            # remove vanished droplets
            if self.keep_vanished:
                emulsion = state.agents.droplets.copy()
            else:
                emulsion = state.agents.droplets.copy(min_radius=0)
            # add emulsion without an additional copy
            self.emulsions.append(emulsion, time=t, copy=False)

        # handle droplet track list
        if self.store_droplet_tracks is not False:
            for i, droplet in enumerate(state.agents.droplets):
                track = self.droplet_tracks[i]
                is_active = track and track.last.radius > 0
                if is_active or droplet.radius > 0 or self.keep_vanished:
                    track.append(droplet, time=t)
    
    
    def finalize(self, info: InfoDict = None) -> None:
        """ finalize the tracker, supplying additional information

        Args:
            info (dict):
                Extra information from the simulation        
        """
        super().finalize(info)
        # write data to files, if requested
        if isinstance(self.store_emulsions, str):
            self.emulsions.to_file(self.store_emulsions)
        if isinstance(self.store_droplet_tracks, str):
            self.droplet_tracks.to_file(self.store_droplet_tracks)
        
        